@@ -50,17 +50,17 @@
 G_DEFINE_TYPE (TfStream, tf_stream, G_TYPE_OBJECT);
 
 #define DEBUG(stream, format, ...) \
-  g_debug ("stream %d (%s) %s: " format, \
-    stream->stream_id, \
-    (stream->priv->media_type == FS_MEDIA_TYPE_AUDIO) ? "audio" \
+  g_debug ("stream %d %p (%s) %s: " format, \
+    stream->stream_id, stream,                                   \
+    (stream->priv->media_type == FS_MEDIA_TYPE_AUDIO) ? "audio"  \
                                                       : "video", \
     G_STRFUNC, \
     ##__VA_ARGS__)
 
 #define WARNING(stream, format, ...) \
-  g_warning ("stream %d (%s) %s: " format, \
-    stream->stream_id, \
-    (stream->priv->media_type == FS_MEDIA_TYPE_AUDIO) ? "audio" \
+  g_warning ("stream %d %p (%s) %s: " format, \
+    stream->stream_id, stream,                                   \
+    (stream->priv->media_type == FS_MEDIA_TYPE_AUDIO) ? "audio"  \
                                                       : "video", \
     G_STRFUNC, \
     ##__VA_ARGS__)
@@ -79,8 +79,6 @@ struct _TfStreamPrivate
   const TfNatProperties *nat_props;
   GList *local_preferences;
 
-  GError *construction_error;
-
   TpMediaStreamHandler *stream_handler_proxy;
 
   FsStreamDirection desired_direction;
@@ -89,10 +87,12 @@ struct _TfStreamPrivate
 
   GList *local_candidates;
 
+  GList *last_sent_codecs;
+
   gboolean send_local_codecs;
   gboolean send_supported_codecs;
 
-  gboolean gathering;
+  NewStreamCreatedCb *new_stream_created_cb;
 };
 
 enum
@@ -122,7 +122,11 @@ enum
   PROP_LOCAL_PREFERENCES
 };
 
-
+static void get_all_properties_cb (TpProxy *proxy,
+    GHashTable *out_Properties,
+    const GError *error,
+    gpointer user_data,
+    GObject *weak_object);
 static gboolean tf_stream_request_resource (
     TfStream *self,
     TpMediaStreamDirection dir);
@@ -175,6 +179,8 @@ static void async_method_callback (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     const GError *error,
     gpointer user_data,
     GObject *weak_object);
+static void tf_stream_shutdown (TfStream *self);
+
 
 static void cb_fs_stream_src_pad_added (FsStream *fsstream G_GNUC_UNUSED,
     GstPad *pad,
@@ -295,153 +301,19 @@ tf_stream_constructor (GType type,
 {
   GObject *obj;
   TfStream *stream;
-  TfStreamPrivate *priv;
-  TfStreamClass *klass = NULL;
-  gchar *transmitter;
-  guint n_args = 0;
-  GList *preferred_local_candidates = NULL;
-  GParameter params[MAX_STREAM_TRANS_PARAMS];
+
 
   obj = G_OBJECT_CLASS (tf_stream_parent_class)->
-            constructor (type, n_props, props);
-  stream = (TfStream *) obj;
-  priv = stream->priv;
-  klass = TF_STREAM_GET_CLASS(obj);
+      constructor (type, n_props, props);
 
-  g_signal_connect (priv->stream_handler_proxy, "invalidated",
+  stream = TF_STREAM (obj);
+
+  g_signal_connect (stream->priv->stream_handler_proxy, "invalidated",
       G_CALLBACK (invalidated_cb), obj);
 
-  tp_cli_media_stream_handler_connect_to_add_remote_candidate
-      (priv->stream_handler_proxy, add_remote_candidate, NULL, NULL, obj,
-       NULL);
-  tp_cli_media_stream_handler_connect_to_remove_remote_candidate
-      (priv->stream_handler_proxy, remove_remote_candidate, NULL, NULL, obj,
-       NULL);
-  tp_cli_media_stream_handler_connect_to_set_active_candidate_pair
-      (priv->stream_handler_proxy, set_active_candidate_pair, NULL, NULL, obj,
-       NULL);
-  tp_cli_media_stream_handler_connect_to_set_remote_candidate_list
-      (priv->stream_handler_proxy, set_remote_candidate_list, NULL, NULL, obj,
-       NULL);
-  tp_cli_media_stream_handler_connect_to_set_remote_codecs
-      (priv->stream_handler_proxy, set_remote_codecs, NULL, NULL, obj, NULL);
-  tp_cli_media_stream_handler_connect_to_set_stream_playing
-      (priv->stream_handler_proxy, set_stream_playing, NULL, NULL, obj, NULL);
-  tp_cli_media_stream_handler_connect_to_set_stream_sending
-      (priv->stream_handler_proxy, set_stream_sending, NULL, NULL, obj, NULL);
-  tp_cli_media_stream_handler_connect_to_set_stream_held
-      (priv->stream_handler_proxy, set_stream_held, NULL, NULL, obj, NULL);
-  tp_cli_media_stream_handler_connect_to_start_telephony_event
-      (priv->stream_handler_proxy, start_telephony_event, NULL, NULL, obj,
-       NULL);
-  tp_cli_media_stream_handler_connect_to_stop_telephony_event
-      (priv->stream_handler_proxy, stop_telephony_event, NULL, NULL, obj,
-       NULL);
-  tp_cli_media_stream_handler_connect_to_close
-      (priv->stream_handler_proxy, close, NULL, NULL, obj, NULL);
-
-  memset (params, 0, sizeof(GParameter) * MAX_STREAM_TRANS_PARAMS);
-
-  if (stream->priv->nat_props == NULL ||
-      stream->priv->nat_props->nat_traversal == NULL ||
-      !strcmp (stream->priv->nat_props->nat_traversal, "gtalk-p2p"))
-    {
-      transmitter = "nice";
-
-      params[n_args].name = "compatibility-mode";
-      g_value_init (&params[n_args].value, G_TYPE_UINT);
-      g_value_set_uint (&params[n_args].value, 1);
-      n_args++;
-    }
-  else
-    {
-      transmitter = "rawudp";
-
-      if (stream->priv->media_type == TP_MEDIA_STREAM_TYPE_AUDIO)
-        preferred_local_candidates = g_list_prepend (NULL,
-            fs_candidate_new (NULL, FS_COMPONENT_RTP, FS_CANDIDATE_TYPE_HOST,
-                FS_NETWORK_PROTOCOL_UDP, NULL, 7078));
-      else if (stream->priv->media_type == TP_MEDIA_STREAM_TYPE_VIDEO)
-        preferred_local_candidates = g_list_prepend (NULL,
-            fs_candidate_new (NULL, FS_COMPONENT_RTP, FS_CANDIDATE_TYPE_HOST,
-                FS_NETWORK_PROTOCOL_UDP, NULL, 9078));
-    }
-
-  if (stream->priv->nat_props &&
-      stream->priv->nat_props->stun_server &&
-      stream->priv->nat_props->stun_port)
-    {
-      gchar *conn_timeout_str = NULL;
-
-      params[n_args].name = "stun-ip";
-      g_value_init (&params[n_args].value, G_TYPE_STRING);
-      g_value_set_string (&params[n_args].value,
-          stream->priv->nat_props->stun_server);
-      n_args++;
-
-      params[n_args].name = "stun-port";
-      g_value_init (&params[n_args].value, G_TYPE_UINT);
-      g_value_set_uint (&params[n_args].value,
-          stream->priv->nat_props->stun_port);
-      n_args++;
-
-      conn_timeout_str = getenv ("FS_CONN_TIMEOUT");
-
-      if (conn_timeout_str)
-        {
-          gint conn_timeout = strtol (conn_timeout_str, NULL, 10);
-
-          params[n_args].name = "stun-timeout";
-          g_value_init (&params[n_args].value, G_TYPE_UINT);
-          g_value_set_uint (&params[n_args].value, conn_timeout);
-          n_args++;
-        }
-    }
-
-  if (preferred_local_candidates)
-    {
-      params[n_args].name = "preferred-local-candidates";
-      g_value_init (&params[n_args].value, FS_TYPE_CANDIDATE_LIST);
-      g_value_take_boxed (&params[n_args].value,
-          preferred_local_candidates);
-      n_args++;
-    }
-
-  stream->priv->fs_session = fs_conference_new_session (
-      stream->priv->fs_conference,
-      tp_media_type_to_fs (stream->priv->media_type),
-      &stream->priv->construction_error);
-
-  if (!stream->priv->fs_session)
-    return obj;
-
-  stream->priv->fs_stream = fs_session_new_stream (stream->priv->fs_session,
-      stream->priv->fs_participant,
-      FS_DIRECTION_NONE,
-      transmitter,
-      n_args,
-      params,
-      &stream->priv->construction_error);
-
-  if (!stream->priv->fs_stream)
-    return obj;
-
-  if (stream->priv->local_preferences)
-    if (!fs_session_set_codec_preferences (stream->priv->fs_session,
-            stream->priv->local_preferences,
-            &stream->priv->construction_error))
-      return obj;
-
-
-  if (g_object_class_find_property (
-          G_OBJECT_GET_CLASS (stream->priv->fs_session),
-          "no-rtcp-timeout"))
-    g_object_set (stream->priv->fs_session, "no-rtcp-timeout", 0, NULL);
-
-  g_signal_connect (stream->priv->fs_stream, "src-pad-added",
-      G_CALLBACK (cb_fs_stream_src_pad_added), stream);
-
-  stream->priv->send_local_codecs = TRUE;
+  tp_cli_dbus_properties_call_get_all (stream->priv->stream_handler_proxy,
+      -1, "org.freedesktop.Telepathy.Media.StreamHandler",
+      get_all_properties_cb, NULL, NULL, obj);
 
   return obj;
 }
@@ -488,6 +360,12 @@ tf_stream_dispose (GObject *object)
       priv->local_preferences = NULL;
     }
 
+  if (priv->last_sent_codecs)
+    {
+      fs_codec_list_destroy (priv->last_sent_codecs);
+      priv->last_sent_codecs = NULL;
+    }
+
   fs_candidate_list_destroy (priv->local_candidates);
   priv->local_candidates = NULL;
 
@@ -697,6 +575,315 @@ tf_stream_class_init (TfStreamClass *klass)
           G_TYPE_NONE, 2, GST_TYPE_PAD, FS_TYPE_CODEC);
 }
 
+static void
+get_all_properties_cb (TpProxy *proxy,
+    GHashTable *out_Properties,
+    const GError *dbus_error,
+    gpointer user_data,
+    GObject *weak_object)
+{
+  TfStream *stream = TF_STREAM (weak_object);
+  GError *myerror = NULL;
+  gchar *transmitter;
+  guint n_args = 0;
+  GList *preferred_local_candidates = NULL;
+  GParameter params[MAX_STREAM_TRANS_PARAMS];
+  const gchar *nat_traversal = NULL;
+  GPtrArray *stun_servers;
+  gboolean got_stun = FALSE;
+  GPtrArray *dbus_relay_info;
+
+  if (dbus_error)
+    {
+      tf_stream_error (stream, 0, dbus_error->message);
+      return;
+    }
+
+  tp_cli_media_stream_handler_connect_to_add_remote_candidate
+      (stream->priv->stream_handler_proxy, add_remote_candidate, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_remove_remote_candidate
+      (stream->priv->stream_handler_proxy, remove_remote_candidate, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_active_candidate_pair
+      (stream->priv->stream_handler_proxy, set_active_candidate_pair, NULL,
+          NULL, (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_remote_candidate_list
+      (stream->priv->stream_handler_proxy, set_remote_candidate_list, NULL,
+          NULL, (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_remote_codecs
+      (stream->priv->stream_handler_proxy, set_remote_codecs, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_stream_playing
+      (stream->priv->stream_handler_proxy, set_stream_playing, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_stream_sending
+      (stream->priv->stream_handler_proxy, set_stream_sending, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_stream_held
+      (stream->priv->stream_handler_proxy, set_stream_held, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_start_telephony_event
+      (stream->priv->stream_handler_proxy, start_telephony_event, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_stop_telephony_event
+      (stream->priv->stream_handler_proxy, stop_telephony_event, NULL, NULL,
+          (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_close
+      (stream->priv->stream_handler_proxy, close, NULL, NULL,
+          (GObject*) stream, NULL);
+
+  memset (params, 0, sizeof(GParameter) * MAX_STREAM_TRANS_PARAMS);
+
+  nat_traversal = tp_asv_get_string (out_Properties, "NATTraversal");
+  if (!nat_traversal && stream->priv->nat_props)
+    nat_traversal = stream->priv->nat_props->nat_traversal;
+
+  if (!nat_traversal || !strcmp (nat_traversal, "gtalk-p2p"))
+    {
+      transmitter = "nice";
+
+      params[n_args].name = "compatibility-mode";
+      g_value_init (&params[n_args].value, G_TYPE_UINT);
+      g_value_set_uint (&params[n_args].value, 1);
+      n_args++;
+    }
+  else if (!strcmp (nat_traversal, "ice-udp"))
+    {
+      transmitter = "nice";
+    }
+  else if (!strcmp (nat_traversal, "wlm-8.5"))
+    {
+      transmitter = "nice";
+
+      params[n_args].name = "compatibility-mode";
+      g_value_init (&params[n_args].value, G_TYPE_UINT);
+      g_value_set_uint (&params[n_args].value, 2);
+      n_args++;
+    }
+  else if (!strcmp (nat_traversal, "wlm-2009"))
+    {
+      transmitter = "nice";
+
+      params[n_args].name = "compatibility-mode";
+      g_value_init (&params[n_args].value, G_TYPE_UINT);
+      g_value_set_uint (&params[n_args].value, 3);
+      n_args++;
+    }
+  else
+    {
+      transmitter = "rawudp";
+
+      if (stream->priv->media_type == TP_MEDIA_STREAM_TYPE_AUDIO)
+        preferred_local_candidates = g_list_prepend (NULL,
+            fs_candidate_new (NULL, FS_COMPONENT_RTP, FS_CANDIDATE_TYPE_HOST,
+                FS_NETWORK_PROTOCOL_UDP, NULL, 7078));
+      else if (stream->priv->media_type == TP_MEDIA_STREAM_TYPE_VIDEO)
+        preferred_local_candidates = g_list_prepend (NULL,
+            fs_candidate_new (NULL, FS_COMPONENT_RTP, FS_CANDIDATE_TYPE_HOST,
+                FS_NETWORK_PROTOCOL_UDP, NULL, 9078));
+    }
+
+  /* FIXME: use correct macro when available */
+  stun_servers = tp_asv_get_boxed (out_Properties, "STUNServers",
+      tp_type_dbus_array_su ());
+
+  if (stun_servers && stun_servers->len)
+    {
+      GValueArray *stun_server = g_ptr_array_index (stun_servers, 0);
+
+      if (stun_server && stun_server->n_values == 2)
+        {
+          GValue *stun_ip = g_value_array_get_nth (stun_server, 0);
+          GValue *stun_port = g_value_array_get_nth (stun_server, 1);
+
+          DEBUG (stream, "Adding STUN server %s:%u",
+              g_value_get_string (stun_ip),
+              g_value_get_uint (stun_port));
+
+          params[n_args].name = "stun-ip";
+          g_value_init (&params[n_args].value, G_TYPE_STRING);
+          g_value_copy (stun_ip, &params[n_args].value);
+          n_args++;
+
+          params[n_args].name = "stun-port";
+          g_value_init (&params[n_args].value, G_TYPE_UINT);
+          g_value_copy (stun_port, &params[n_args].value);
+          n_args++;
+
+          got_stun = TRUE;
+        }
+    }
+
+  if (!got_stun && stream->priv->nat_props &&
+      stream->priv->nat_props->stun_server &&
+      stream->priv->nat_props->stun_port)
+    {
+      DEBUG (stream, "Adding STUN server (old API) %s:%u",
+          stream->priv->nat_props->stun_server,
+          stream->priv->nat_props->stun_port);
+      params[n_args].name = "stun-ip";
+      g_value_init (&params[n_args].value, G_TYPE_STRING);
+      g_value_set_string (&params[n_args].value,
+          stream->priv->nat_props->stun_server);
+      n_args++;
+
+      params[n_args].name = "stun-port";
+      g_value_init (&params[n_args].value, G_TYPE_UINT);
+      g_value_set_uint (&params[n_args].value,
+          stream->priv->nat_props->stun_port);
+      n_args++;
+
+      got_stun = TRUE;
+    }
+
+  if (got_stun)
+    {
+      gchar *conn_timeout_str = NULL;
+
+      conn_timeout_str = getenv ("FS_CONN_TIMEOUT");
+      if (conn_timeout_str)
+        {
+          gint conn_timeout = strtol (conn_timeout_str, NULL, 10);
+
+          params[n_args].name = "stun-timeout";
+          g_value_init (&params[n_args].value, G_TYPE_UINT);
+          g_value_set_uint (&params[n_args].value, conn_timeout);
+          n_args++;
+        }
+    }
+
+  dbus_relay_info = tp_asv_get_boxed (out_Properties, "RelayInfo",
+      TP_ARRAY_TYPE_STRING_VARIANT_MAP_LIST);
+
+  if (dbus_relay_info && dbus_relay_info->len)
+    {
+      GValueArray *fs_relay_info = g_value_array_new (0);
+      guint i;
+      GValue val = {0};
+      g_value_init (&val, GST_TYPE_STRUCTURE);
+
+      for (i = 0; i < dbus_relay_info->len; i++)
+        {
+          GHashTable *one_relay = g_ptr_array_index(dbus_relay_info, i);
+          const gchar *type;
+          const gchar *ip;
+          guint32 port;
+          const gchar *username;
+          const gchar *password;
+          guint component;
+          GstStructure *s;
+
+          ip = tp_asv_get_string (one_relay, "ip");
+          port = tp_asv_get_uint32 (one_relay, "port", NULL);
+          type = tp_asv_get_string (one_relay, "type");
+          username = tp_asv_get_string (one_relay, "username");
+          password = tp_asv_get_string (one_relay, "password");
+          component = tp_asv_get_uint32 (one_relay, "component", NULL);
+
+          if (!ip || !port || !username || !password)
+              continue;
+
+          s = gst_structure_new ("relay-info",
+              "ip", G_TYPE_STRING, ip,
+              "port", G_TYPE_UINT, port,
+              "username", G_TYPE_STRING, username,
+              "password", G_TYPE_STRING, password,
+              NULL);
+
+          if (type)
+            gst_structure_set (s, "relay-type", G_TYPE_STRING, type, NULL);
+
+          if (component)
+            gst_structure_set (s, "component", G_TYPE_UINT, component, NULL);
+
+          if (!type)
+            type = "udp";
+
+          DEBUG (stream, "Adding relay (%s) %s:%u %s:%s %u",
+              type, ip, port, username, password, component);
+
+          g_value_take_boxed (&val, s);
+
+          g_value_array_append (fs_relay_info, &val);
+          g_value_reset (&val);
+        }
+
+      if (fs_relay_info->n_values)
+        {
+          params[n_args].name = "relay-info";
+          g_value_init (&params[n_args].value, G_TYPE_VALUE_ARRAY);
+          g_value_set_boxed (&params[n_args].value, fs_relay_info);
+          n_args++;
+        }
+
+      g_value_array_free (fs_relay_info);
+    }
+
+  if (preferred_local_candidates)
+    {
+      params[n_args].name = "preferred-local-candidates";
+      g_value_init (&params[n_args].value, FS_TYPE_CANDIDATE_LIST);
+      g_value_take_boxed (&params[n_args].value,
+          preferred_local_candidates);
+      n_args++;
+    }
+
+  stream->priv->fs_session = fs_conference_new_session (
+      stream->priv->fs_conference,
+      tp_media_type_to_fs (stream->priv->media_type),
+      &myerror);
+
+  if (!stream->priv->fs_session)
+    {
+      tf_stream_error (stream, 0, myerror->message);
+      WARNING (stream, "Error creating session: %s", myerror->message);
+      g_clear_error (&myerror);
+      return;
+    }
+
+  stream->priv->fs_stream = fs_session_new_stream (stream->priv->fs_session,
+      stream->priv->fs_participant,
+      FS_DIRECTION_NONE,
+      transmitter,
+      n_args,
+      params,
+      &myerror);
+
+  if (!stream->priv->fs_stream)
+    {
+      tf_stream_error (stream, 0, myerror->message);
+      WARNING (stream, "Error creating stream: %s", myerror->message);
+      g_clear_error (&myerror);
+      return;
+    }
+
+  if (stream->priv->local_preferences)
+    if (!fs_session_set_codec_preferences (stream->priv->fs_session,
+            stream->priv->local_preferences,
+            &myerror))
+      {
+        tf_stream_error (stream, 0, myerror->message);
+        WARNING (stream, "Error setting codec preferences: %s",
+            myerror->message);
+        g_clear_error (&myerror);
+        return;
+      }
+
+
+  if (g_object_class_find_property (
+          G_OBJECT_GET_CLASS (stream->priv->fs_session),
+          "no-rtcp-timeout"))
+    g_object_set (stream->priv->fs_session, "no-rtcp-timeout", 0, NULL);
+
+  g_signal_connect (stream->priv->fs_stream, "src-pad-added",
+      G_CALLBACK (cb_fs_stream_src_pad_added), stream);
+
+  stream->priv->send_local_codecs = TRUE;
+
+  stream->priv->new_stream_created_cb (stream, stream->priv->channel);
+}
+
 /* dummy callback handler for async calling calls with no return values */
 static void
 async_method_callback (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
@@ -818,9 +1005,11 @@ cb_fs_local_candidates_prepared (TfStream *self)
       tp_cli_media_stream_handler_call_new_native_candidate (
           self->priv->stream_handler_proxy, -1, foundation, transports,
           async_method_callback,
-          "Media.StreamHandler::NativeCandidatesPrepared",
+          "Media.StreamHandler::NewNativeCandidate",
           NULL, (GObject *) self);
 
+      g_boxed_free (TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT_LIST,
+          transports);
       g_free (foundation);
     }
 
@@ -1275,10 +1464,6 @@ set_stream_sending (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
   else
     {
       if (!self->priv->held) {
-        if (!self->priv->gathering)
-          tf_stream_free_resource (self,
-              TP_MEDIA_STREAM_DIRECTION_RECEIVE);
-
         g_object_set (self->priv->fs_stream,
             "direction", current_direction & ~(FS_DIRECTION_SEND),
             NULL);
@@ -1371,7 +1556,7 @@ set_stream_held (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
          {
            tp_cli_media_stream_handler_call_hold_state (
              self->priv->stream_handler_proxy, -1, TRUE,
-             async_method_callback, "Media.StreamHandler::HoldState",
+             async_method_callback, "Media.StreamHandler::HoldState TRUE",
              NULL, (GObject *) self);
          }
        self->priv->held = TRUE;
@@ -1380,9 +1565,6 @@ set_stream_held (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     {
       FsStreamDirection desired_direction = self->priv->desired_direction;
 
-      if (self->priv->gathering)
-        desired_direction |= TP_MEDIA_STREAM_DIRECTION_SEND;
-
       if (tf_stream_request_resource (self, desired_direction))
         {
            g_object_set (self->priv->fs_stream,
@@ -1390,7 +1572,7 @@ set_stream_held (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
                NULL);
            tp_cli_media_stream_handler_call_hold_state (
              self->priv->stream_handler_proxy, -1, FALSE,
-             async_method_callback, "Media.StreamHandler::HoldState",
+             async_method_callback, "Media.StreamHandler::HoldState FALSE",
              NULL, (GObject *) self);
 
            self->priv->held = FALSE;
@@ -1437,6 +1619,21 @@ stop_telephony_event (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     WARNING (self, "stopping event failed");
 }
 
+
+static void
+tf_stream_shutdown (TfStream *self)
+{
+  if (self->priv->fs_stream)
+    g_object_set (self->priv->fs_stream,
+        "direction", FS_DIRECTION_NONE,
+        NULL);
+  tf_stream_free_resource (self,
+      TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL);
+
+  g_signal_emit (self, signals[CLOSED], 0);
+}
+
+
 static void
 close (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
        gpointer user_data G_GNUC_UNUSED,
@@ -1446,13 +1643,7 @@ close (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 
   DEBUG (self, "close requested by connection manager");
 
-  g_object_set (self->priv->fs_stream,
-            "direction", FS_DIRECTION_NONE,
-            NULL);
-  tf_stream_free_resource (self,
-      TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL);
-
-  g_signal_emit (self, signals[CLOSED], 0);
+  tf_stream_shutdown (self);
 }
 
 static void
@@ -1495,7 +1686,7 @@ cb_fs_new_active_candidate_pair (TfStream *self,
 
   tp_cli_media_stream_handler_call_stream_state (
     self->priv->stream_handler_proxy, -1, TP_MEDIA_STREAM_STATE_CONNECTED,
-    async_method_callback, "Media.StreamHandler::SetStreamState",
+    async_method_callback, "Media.StreamHandler::StreamState",
     NULL, (GObject *) self);
 }
 
@@ -1508,6 +1699,8 @@ invalidated_cb (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 {
   TfStream *stream = TF_STREAM (user_data);
 
+  DEBUG (stream, "proxy invalidated");
+
   if (stream->priv->stream_handler_proxy)
     {
       TpMediaStreamHandler *tmp = stream->priv->stream_handler_proxy;
@@ -1515,6 +1708,8 @@ invalidated_cb (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
       stream->priv->stream_handler_proxy = NULL;
       g_object_unref (tmp);
     }
+
+  tf_stream_shutdown (stream);
 }
 
 /**
@@ -1563,6 +1758,9 @@ _tf_stream_bus_message (TfStream *stream,
   if (GST_MESSAGE_TYPE (message) != GST_MESSAGE_ELEMENT)
     return FALSE;
 
+  if (!stream->priv->fs_stream || !stream->priv->fs_session)
+    return FALSE;
+
   if (gst_structure_has_name (s, "farsight-error"))
     {
       GObject *object;
@@ -1738,8 +1936,7 @@ _tf_stream_new (gpointer channel,
     TpMediaStreamDirection direction,
     TfNatProperties *nat_props,
     GList *local_preferences,
-    GError **error)
-
+    NewStreamCreatedCb new_stream_created_cb)
 {
   TfStream *self = NULL;
 
@@ -1755,13 +1952,7 @@ _tf_stream_new (gpointer channel,
       "codec-preferences", local_preferences,
       NULL);
 
-  if (self->priv->construction_error)
-    {
-      tf_stream_error (self, 0, self->priv->construction_error->message);
-      g_propagate_error (error, self->priv->construction_error);
-      g_object_unref (self);
-      return NULL;
-    }
+  self->priv->new_stream_created_cb = new_stream_created_cb;
 
   return self;
 }
@@ -1772,47 +1963,24 @@ _tf_stream_try_sending_codecs (TfStream *stream)
   gboolean ready = FALSE;
   GList *fscodecs = NULL;
   GList *item = NULL;
+  GPtrArray *tpcodecs = NULL;
+
+  gboolean sent_codecs = FALSE;
 
   DEBUG (stream, "called (send_local:%d send_supported:%d)",
       stream->priv->send_local_codecs, stream->priv->send_supported_codecs);
 
-  if (!stream->priv->send_supported_codecs && !stream->priv->send_local_codecs)
-    return;
-
   g_object_get (stream->priv->fs_session, "codecs-ready", &ready, NULL);
 
-  if (!ready)
+  if (!ready && stream->priv->has_resource & TP_MEDIA_STREAM_DIRECTION_SEND)
     {
-      if (!stream->priv->gathering)
-        {
-          DEBUG (stream, "Enabling resources to gather codecs");
-          stream->priv->gathering = TRUE;
-
-          if (stream->priv->held)
-            return;
-
-          if (!tf_stream_request_resource (stream,
-                  TP_MEDIA_STREAM_DIRECTION_SEND))
-            {
-              DEBUG (stream, "Could not get resource, sending may not work");
-              stream->priv->gathering = FALSE;
-              goto ignore_ready;
-            }
-        }
+      DEBUG (stream, "Ignoring new codecs because we're sending,"
+          " but we're not ready");
       return;
     }
 
- ignore_ready:
-
   g_object_get (stream->priv->fs_session, "codecs", &fscodecs, NULL);
 
-  if (stream->priv->gathering)
-    {
-      if (!(stream->priv->desired_direction & TP_MEDIA_STREAM_DIRECTION_SEND))
-        tf_stream_free_resource (stream, TP_MEDIA_STREAM_DIRECTION_SEND);
-      stream->priv->gathering = FALSE;
-    }
-
   for(item = fscodecs; item; item = g_list_next (item))
     {
       gchar *tmp = fs_codec_to_string (item->data);
@@ -1822,7 +1990,7 @@ _tf_stream_try_sending_codecs (TfStream *stream)
 
   if (stream->priv->send_local_codecs)
     {
-      GPtrArray *tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+      tpcodecs = fs_codecs_to_tp (stream, fscodecs);
 
       DEBUG (stream, "calling MediaStreamHandler::Ready");
       tp_cli_media_stream_handler_call_ready (
@@ -1830,11 +1998,12 @@ _tf_stream_try_sending_codecs (TfStream *stream)
           -1, tpcodecs, async_method_callback, "Media.StreamHandler::Ready",
           NULL, (GObject *) stream);
       stream->priv->send_local_codecs = FALSE;
+      sent_codecs = TRUE;
     }
 
   if (stream->priv->send_supported_codecs)
     {
-      GPtrArray *tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+      tpcodecs = fs_codecs_to_tp (stream, fscodecs);
 
       DEBUG (stream, "calling MediaStreamHandler::SupportedCodecs");
       tp_cli_media_stream_handler_call_supported_codecs (
@@ -1842,9 +2011,26 @@ _tf_stream_try_sending_codecs (TfStream *stream)
           -1, tpcodecs, async_method_callback,
           "Media.StreamHandler::SupportedCodecs", NULL, (GObject *) stream);
       stream->priv->send_supported_codecs = FALSE;
+      sent_codecs = TRUE;
     }
 
-  fs_codec_list_destroy (fscodecs);
+
+  if (!sent_codecs &&
+      !fs_codec_list_are_equal (fscodecs, stream->priv->last_sent_codecs))
+    {
+      tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+
+      DEBUG (stream, "calling MediaStreamHandler::CodecsUpdated");
+      tp_cli_media_stream_handler_call_codecs_updated (
+          stream->priv->stream_handler_proxy,
+          -1, tpcodecs, async_method_callback,
+          "Media.StreamHandler::CodecsUpdated", NULL, (GObject *) stream);
+    }
+
+  if (tpcodecs)
+    g_boxed_free (TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_CODEC_LIST, tpcodecs);
+  fs_codec_list_destroy (stream->priv->last_sent_codecs);
+  stream->priv->last_sent_codecs = fscodecs;
 }
 
 /**
